def get_max(arr: list) -> int:
    """Returns the maximum value in an list

    Args:
        arr (list): the list to find the value in

    Returns:
        int: the value itself
    """
    max_value = arr[0]
    for value in arr:
        if value > max_value:
            max_value = value
    return max_value


MIN_MERGE = 32

def calc_minrun(size: int):
    """Calculates the minimum size of the runs for a certain number of items.

    Args:
        n (int): the number of items in the array
    """

    r = 0
    while size >= MIN_MERGE:
        r |= size & 1
        size >>= 1
    
    return size + r

def insertion_sort(arr: list, left: int, right: int) -> list:
    """Insertion sort from left index to right index
    Which is at most run_size

    Args:
        arr (list): the array to sort
        left (int): the left-most index to start from
        right (int): the right-most index to end from

    Returns:
        list: the sorted list
    """
    for i in range(left + 1, right + 1):
        j = i

        while j > left and arr[j] < arr[j - 1]:
            arr[j], arr[j - 1] = arr[j - 1], arr[j]
            j -= 1


def merge(arr: list, left: int, middle: int, right: int):
    """Merge sorted runs

    Args:
        arr (list): the list to be merged
        left (int): the left index to 
        middle (int): _description_
        right (int): _description_
    """
    len1, len2 = middle - left + 1, right - middle
    left_arr, right_arr = [], []

    left_arr.extend(arr[left + i] for i in range(len1))
    right_arr.extend(arr[middle + i + 1] for i in range(len2))

    l_index, r_index, k = 0, 0, left

    while l_index < len1 and r_index < len2:
        if left[l_index] <= right[r_index]:
            arr[k] = left[l_index]
            l_index += 1
        else:
            arr[k] = right[r_index]
            r_index += 1
    
        k += 1
    
    while l_index < len1:
        arr[k] = left[l_index]
        k += 1
        l_index += 1
    
    while r_index < len2:
        arr[k] = right[r_index]
        k += 1
        r_index += 1


def timsort(arr: list):
    size = len(arr)

    min_run = calc_minrun(size)

    for start in range(0, size, min_run):
        end = min(start + min_run - 1, size - 1)
        insertion_sort(arr, start, end)

    run_size = min_run
    while run_size < size:
        for left in range(0, size, 2 * size):
            pass


def merge(arr1: list, arr2: list) -> list:
    """Merge two arrays for merge sort

    Args:
        arr1 (list): the first array
        arr2 (list): the second array

    Returns:
        list: the merged array
    """
    merged = []

    # Compare elements for both arrays
    while arr1 and arr2:
        if (arr1[0] > arr2[0]):
            merged.append(arr2[0])
            arr2.pop(0)
        else:
            merged.append(arr1[0])
            arr1.pop(0)
    
    # Either arr1 or arr2 is empty now
    # Empty arr1
    while arr1:
        merged.append(arr1[0])
        arr1.pop(0)
    
    # Empty arr2
    while arr2:
        merged.append(arr2[0])
        arr2.pop(0)
    
    return merged

def merge_sort(arr: list) -> list:
    """Sort a list using merge sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    size = len(arr)

    # End condition for recursion
    if (size == 1):
        return arr

    # Get the index to split the array in half
    index = size // 2

    # Get the two split arrays
    arr1 = arr[:index]
    arr2 = arr[index:]

    # Recursivly call merge sort
    arr1 = merge_sort(arr1)
    arr2 = merge_sort(arr2)

    # Merge the two arrays back together
    return merge(arr1, arr2)



def partition(arr: list, low: int, high: int) -> int:
    """Partition the array

    Args:
        arr (list): the array to partition
        low (int): the left-most index
        high (int): the right-most index

    Returns:
        int: the new pivot location
    """
    i = low - 1
    pivot = arr[high]

    # Swap the numbers of the right-most and left-most index
    for j in range(low, high):
        if arr[j] <= pivot:
            i += 1
            arr[i], arr[j] = arr[j], arr[i]
    
    # Swap the right-most number with the pivot
    arr[i+1], arr[high] = arr[high], arr[i+1]

    return i + 1

def quick_sort(arr: list, low: int, high: int):
    if len(arr) == 1:
        return arr
    
    # If there is an array to be sorted
    if low < high:
        # Partition the array
        pivot_location = partition(arr, low, high)

        # Quicksort the partitions
        quick_sort(arr, low, pivot_location - 1)
        quick_sort(arr, pivot_location + 1, high)


from profq_sorting.sorting.bubble_sort import bubble_sort
from profq_sorting.sorting.counting_sort import counting_sort
from profq_sorting.sorting.insertion_sort import insertion_sort
from profq_sorting.sorting.merge_sort import merge_sort
from profq_sorting.sorting.quick_sort import quick_sort
from profq_sorting.sorting.radix_sort import radix_sort
from profq_sorting.sorting.selection_sort import selection_sort
from profq_sorting.sorting.shell_sort import shell_sort
from profq_sorting.sorting.timsort import timsort